from collections import defaultdict

from fnvhash import fnv1a_32
from sqlalchemy.exc import IntegrityError

__all__ = ['EmmaaDatabaseManager', 'EmmaaDatabaseError']

import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .schema import EmmaaTable, User, Query, Base, Result, UserQuery
from emmaa.queries import Query as QueryObject

logger = logging.getLogger(__name__)


class EmmaaDatabaseError(Exception):
    pass


class EmmaaDatabaseSessionManager(object):
    """A Database session context manager that is used by EmmaaDatabaseManager.
    """
    def __init__(self, host, engine):
        logger.debug(f"Grabbing a session to {host}...")
        DBSession = sessionmaker(bind=engine)
        logger.debug("Session grabbed.")
        self.session = DBSession()
        if self.session is None:
            raise EmmaaDatabaseError("Could not acquire session.")
        return

    def __enter__(self):
        return self.session

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type:
            logger.exception(exception_value)
            logger.info("Got exception: rolling back.")
            self.session.rollback()
        else:
            logger.debug("Committing changes...")
            self.session.commit()

        # Close the session.
        self.session.close()


class EmmaaDatabaseManager(object):
    """A class used to manage sessions with EMMAA's database."""
    table_order = ['user', 'query', 'user_query', 'result']

    def __init__(self, host, label=None):
        self.host = host
        self.label = label
        self.engine = create_engine(host)
        self.tables = {tbl.__tablename__: tbl
                       for tbl in EmmaaTable.__subclasses__()}
        self.session = None
        return

    def get_session(self):
        return EmmaaDatabaseSessionManager(self.host, self.engine)

    def create_tables(self, tables=None):
        """Create the tables from the EMMAA database

        Optionally specify `tables` to be created. List may contain either
        table objects or the string names of the tables.
        """
        # Regularize the type of input to name strings.
        if tables is not None:
            tables = [tbl.__tablename__ if isinstance(tbl, EmmaaTable) else tbl
                      for tbl in tables]

        if tables is None:
            tables = set(self.tables.keys())
        else:
            tables = set(tables)

        for tbl_name in self.table_order:
            if tbl_name in tables:
                logger.info(f"Creating {tbl_name} table")
                if not self.tables[tbl_name].__table__.exists(self.engine):
                    self.tables[tbl_name].__table__.create(bind=self.engine)
                    logger.debug("Table created!")
                else:
                    logger.warning(f"Table {tbl_name} already exists! "
                                   f"No action taken.")
        return

    def drop_tables(self, tables=None, force=False):
        """Drop the tables from the EMMAA database given in `tables`.

        If `tables` is None, all tables will be dropped. Note that if `force`
        is False, a warning prompt will be raised to asking for confirmation,
        as this action will remove all data from that table.
        """
        # Regularize the type of input to table objects.
        if tables is not None:
            tables = [tbl if isinstance(tbl, EmmaaTable) else self.tables[tbl]
                      for tbl in tables]

        if not force:
            # Build the message
            if tables is None:
                msg = ("Do you really want to clear the %s database? [y/N]: "
                       % self.label)
            else:
                msg = "You are going to clear the following tables:\n"
                msg += '\n'.join(['\t-' + tbl.__tablename__ for tbl in tables])
                msg += '\n'
                msg += ("Do you really want to clear these tables from %s? "
                        "[y/N]: " % self.label)

            # Check to make sure.
            resp = input(msg)
            if resp != 'y' and resp != 'yes':
                logger.info('Aborting drop.')
                return False

        if tables is None:
            logger.info("Removing all tables...")
            Base.metadata.drop_all(self.engine)
            logger.debug("All tables removed.")
        else:
            for tbl in tables:
                logger.info("Removing %s..." % tbl.__tablename__)
                if tbl.__table__.exists(self.engine):
                    tbl.__table__.drop(self.engine)
                    logger.debug("Table removed.")
                else:
                    logger.debug("Table doesn't exist.")
        return True

    def add_user(self, user_id, email):
        """Add a new user's email and id to Emmaa's User table."""
        try:
            new_user = User(id=user_id, email=email)
            with self.get_session() as sess:
                sess.add(new_user)
        except IntegrityError as e:
            logger.warning(f"A user with email {email} already exists.")
        return user_id

    def put_queries(self, user_email, user_id, query, model_ids,
                    subscribe=True):
        """Add queries to the database for a given user.

        Parameters
        ----------
        user_email : str
            the email of the user that entered the queries.
        user_id : int
            the user id of the user that entered the queries. Corresponds to
            the user id in the User table in indralab_auth_tools
        query : emmaa.queries.Query
            A query object containing all necessary information.
        model_ids : list[str]
            A list of the short, standard model IDs to which the user wishes
            to apply these queries.
        subscribe : bool
            True if the user wishes to subscribe to this query.
        """
        logger.info(f"Got request to put query {query} for {user_email} "
                    f"for {model_ids} with subscribe={subscribe}")

        # Make sure model_ids is a list.
        if not isinstance(model_ids, list) and not isinstance(model_ids, set):
            raise TypeError("Invalid type: %s. Must be list or set."
                            % type(model_ids))

        if not subscribe:
            logger.info("Not subscribing...")

        # Check if anonymous user
        if not user_email and not user_id:
            logger.info(f'User {user_email} is not registered in the user '
                        'database. Query will be stored as anonymous query.')
            # Make sure user_id is a None and not any other object
            # evaluating to False (only None register as NULL in table)
            user_email = 'anonymous@emmaa.bio'
            user_id = None

        # Open database session
        with self.get_session() as sess:
            # Get existing hashes, user's id and user's subscriptions
            existing_hashes = {h for h, in sess.query(Query.hash).all()}
            existing_user_queries = {h for h, in sess.query(
                UserQuery.query_hash).filter(UserQuery.user_id == user_id)}

            # Check if logged in user is in the emmaa user table
            if user_email and user_id:
                res = \
                    sess.query(User.id).filter(User.id == user_id).first()
                if res:
                    logger.info(f'User {user_email} is registered in the '
                                f'user table.')
                else:
                    logger.info(f'{user_email} not in user table. Adding...')
                    self.add_user(user_id=user_id, email=user_email)

            new_queries = []
            new_user_queries = []
            for model_id in model_ids:
                qh = query.get_hash_with_model(model_id)

                # Add to queries if not present
                if qh not in existing_hashes:
                    logger.info(f"Adding query on {model_id} to the db.")
                    new_queries.append(Query(model_id=model_id,
                                             json=query.to_json(),
                                             hash=qh))
                else:
                    logger.info(f"Query for {model_id} already in db.")

                # Add query to UserQuery table or update existing one
                if qh not in existing_user_queries:
                    new_user_queries.append(UserQuery(user_id=user_id,
                                                      query_hash=qh,
                                                      subscription=subscribe,
                                                      count=1))
                    logger.info(f'Registering query on {model_id} for user '
                                f'{user_email}')
                # Update existing query
                else:
                    user_query = sess.query(UserQuery).filter(
                        UserQuery.user_id == user_id,
                        UserQuery.query_hash == qh
                    ).first()
                    logger.info(f'Updating existing query for {user_email} '
                                f'on {model_id} ({qh})')
                    # Update subscription
                    # Set subscribe to True, handle un-subscribe elsewhere
                    if subscribe:
                        user_query = update_subscription(user_query, subscribe)

                    # Update query count
                    user_query.count += 1

            # Add new queries and register them for the user
            sess.add_all(new_queries)
            sess.add_all(new_user_queries)
        return

    def get_queries(self, model_id):
        """Get queries that refer to the given model_id.

        Parameters
        ----------
        model_id : str
            The short, standard model ID.

        Returns
        -------
        queries : list[emmaa.queries.Query]
            A list of queries retrieved from the database.
        """
        # TODO: check whether a query is registered or not.
        with self.get_session() as sess:
            q = sess.query(Query.json).filter(Query.model_id == model_id)
            queries = [QueryObject._from_json(q) for q, in q.all()]
        return queries

    def put_results(self, model_id, query_results):
        """Add new results for a set of queries tested on a model_id.

        Parameters
        ----------
        model_id : str
            The short, standard model ID.
        query_results : list of tuples
            A list of tuples of the form (query, result_json), where
            the query is the query object run against the model,
            and the result_json is the json containing corresponding result.
        """
        results = []
        for query, mc_type, result_json in query_results:
            query_hash = query.get_hash_with_model(model_id)
            results.append(Result(query_hash=query_hash,
                                  mc_type=mc_type,
                                  result_json=result_json))

        with self.get_session() as sess:
            sess.add_all(results)
        return

    def get_results_from_query(self, query, model_ids, latest_order=1):
        logger.info(f"Got request for results of {query} on {model_ids}.")
        hashes = {query.get_hash_with_model(model_id) for model_id in model_ids}
        with self.get_session() as sess:
            q = (sess.query(Query.model_id, Query.json, Result.mc_type,
                            Result.result_json, Result.date)
                 .filter(Result.query_hash.in_(hashes),
                         Query.hash == Result.query_hash))
            results = _make_queries_in_results(q.all())
            results = _weed_results(results, latest_order=latest_order)
        logger.info(f"Found {len(results)} results.")
        return results

    def get_results(self, user_email, latest_order=1):
        """Get the results for which the user has registered.

        Parameters
        ----------
        user_email : str
            The email of a user.

        Returns
        -------
        results : list[tuple]
            A list of tuples, each of the form: (model_id, query, mc_type,
            result_json, date) representing the result of a query run on a
            model on a given date.
        """
        logger.info(f"Got request for results for {user_email}")
        with self.get_session() as sess:
            q = (sess.query(Query.model_id, Query.json, Result.mc_type,
                            Result.result_json, Result.date)
                 .filter(Query.hash == Result.query_hash,
                         Query.hash == UserQuery.query_hash,
                         UserQuery.user_id == User.id,
                         UserQuery.subscription,
                         User.email == user_email))
            results = _make_queries_in_results(q.all())
            results = _weed_results(results, latest_order=latest_order)
        logger.info(f"Found {len(results)} results.")
        return results

    def get_users(self, query, model_id):
        logger.info(f"Got request for users for {query} in {model_id}.")
        with self.get_session() as sess:
            q = (sess.query(User.email).filter(
                    User.id == UserQuery.user_id,
                    Query.hash == query.get_hash_with_model(model_id)))
            users = [q for q, in q.all()]
        return users


def _weed_results(result_iter, latest_order=1):
    # Each element of result_iter: (model_id, query(object), result_json, date)
    result_dict = defaultdict(list)
    for res in result_iter:
        result_dict[(res[1].get_hash_with_model(res[0]), res[2])].append(
            tuple(res))
    sorted_results = [sorted(res_list, key=lambda r: r[-1])
                      for res_list in result_dict.values()]
    results = [result[-latest_order] for result in sorted_results]
    return results


def _make_queries_in_results(result_iter):
    # Each element of result_iter: (model_id, query_json, result_json, date)
    # Replace query_json with Query object
    results = []
    for res in result_iter:
        query = QueryObject._from_json(res[1])
        results.append((res[0], query, res[2], res[3], res[4]))
    return results


def sorted_json_string(json_thing):
    """Produce a string that is unique to a json's contents."""
    if isinstance(json_thing, str):
        return json_thing
    elif isinstance(json_thing, list):
        return '[%s]' % (','.join(sorted(sorted_json_string(s)
                                         for s in json_thing)))
    elif isinstance(json_thing, dict):
        return '{%s}' % (','.join(sorted(k + sorted_json_string(v)
                                         for k, v in json_thing.items())))
    elif isinstance(json_thing, float):
        return str(json_thing)
    else:
        raise TypeError(f"Invalid type: {type(json_thing)}")


def hash_query(query_json, model_id):
    """Create an FNV-1a 32-bit hash from the query json and model_id."""
    unique_string = model_id + ':' + sorted_json_string(query_json)
    return fnv1a_32(unique_string.encode('utf-8'))


def update_subscription(user_query, new_sub_status):
    """Update a UserQuery object's subscription status

    user_query : `emmaa.db.schema.UserQuery`
        The UserQuery object to be updated
    new_sub_status : Bool
        The subscription status to change to

    Returns
    -------
    user_query : UserQuery(object)
        The updated UserQuery object
    """
    if new_sub_status is not user_query.subscription:
        user_query.subscription = new_sub_status
        logger.info(f'Updated subscription status to '
                    f'{new_sub_status} for query {user_query.query_hash}')
    return user_query



from os import path
from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(name='emmaa',
      version='1.4.0',
      description='Ecosystem of Machine-maintained Models with ' + \
                  'Automated Analysis',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/indralab/emmaa',
      author='EMMAA developers, Harvard Medical School',
      author_email='benjamin_gyori@hms.harvard.edu',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
        ],
      packages=find_packages(),
      install_requires=['indra', 'boto3', 'jsonpickle', 'kappy==4.0.0rc1',
                        'pygraphviz', 'fnvhash', 'sqlalchemy', 'inflection',
                        'pybel', 'flask_jwt_extended'],
      extras_require={'test': ['nose', 'coverage', 'python-coveralls']}
      )



from os.path import abspath, dirname, join
from datetime import datetime
from nose.plugins.attrib import attr
from emmaa.answer_queries import QueryManager, format_results, \
    load_model_manager_from_s3, is_query_result_diff
from emmaa.queries import Query
from emmaa.model_tests import ModelManager
from emmaa.tests.test_db import _get_test_db
from emmaa.model import EmmaaModel


test_query = {'type': 'path_property', 'path': {'type': 'Activation',
              'subj': {'type': 'Agent', 'name': 'BRAF',
                       'db_refs': {'HGNC': '1097'}},
              'obj': {'type': 'Agent', 'name': 'MAPK1',
                      'db_refs': {'HGNC': '6871'}}}}
simple_query = {'typeSelection': 'Activation',
                'subjectSelection': 'BRAF',
                'objectSelection': 'MAPK1'}
query_object = Query._from_json(test_query)

test_response = {3801854542: [
    ('BRAF activates MAP2K1.',
     'https://db.indra.bio/statements/from_agents?subject=1097@HGNC&object='
     '6840@HGNC&type=Activation&format=html'),
    ('Active MAP2K1 activates MAPK1.',
     'https://db.indra.bio/statements/from_agents?subject=6840@HGNC&object='
     '6871@HGNC&type=Activation&format=html')]}
processed_link = '<a href="https://db.indra.bio/statements/from_agents?'\
    'subject=1097@HGNC&object=6840@HGNC&type=Activation&format=html" '\
                 'target="_blank" class="status-link">'\
                 'BRAF activates MAP2K1.</a>'
query_not_appl = {2413475507: [
    ('Query is not applicable for this model',
     'https://emmaa.readthedocs.io/en/latest/dashboard/response_codes.html')]}
# Create a new ModelManager for tests instead of depending on S3 version
test_model = EmmaaModel.load_from_s3('test')
test_mm = ModelManager(test_model)


def test_load_model_manager_from_s3():
    mm = load_model_manager_from_s3('test')
    assert isinstance(mm, ModelManager)


def test_format_results():
    results = [('test', query_object, 'pysb', test_response, datetime.now())]
    formatted_results = format_results(results)
    assert len(formatted_results) == 1
    assert formatted_results[0]['model'] == 'test'
    assert formatted_results[0]['query'] == simple_query
    assert formatted_results[0]['mc_type'] == 'pysb'
    assert isinstance(formatted_results[0]['response'], str)
    assert isinstance(formatted_results[0]['date'], str)


@attr('nonpublic')
def test_answer_immediate_query():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    results = qm.answer_immediate_query('tester@test.com', 1, query_object,
                                        ['test'], subscribe=False)
    assert len(results) == 1
    assert results[0]['model'] == 'test'
    assert results[0]['query'] == simple_query
    assert isinstance(results[0]['response'], str)
    assert 'BRAF activates MAP2K1.' in results[0]['response'], \
        results[0]['response']
    assert isinstance(results[0]['date'], str)


@attr('nonpublic')
def test_answer_get_registered_queries():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    qm.db.put_queries('tester@test.com', 1, query_object, ['test'],
                      subscribe=True)
    qm.answer_registered_queries('test')
    results = qm.get_registered_queries('tester@test.com')
    assert len(results) == 1
    assert results[0]['model'] == 'test'
    assert results[0]['query'] == simple_query
    assert isinstance(results[0]['response'], str)
    assert 'BRAF activates MAP2K1.' in results[0]['response'], \
        results[0]['response']
    assert isinstance(results[0]['date'], str)


def test_is_diff():
    assert not is_query_result_diff(query_not_appl, query_not_appl)
    assert is_query_result_diff(test_response, query_not_appl)


@attr('nonpublic')
def test_report_one_query():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    # Using results from db
    qm.db.put_queries('tester@test.com', 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_results('test', [(query_object, 'pysb', test_response),
                               (query_object, 'pysb', query_not_appl)])
    str_msg = qm.get_report_per_query('test', query_object)[0]
    assert str_msg
    assert 'A new result to query' in str_msg
    assert 'Query is not applicable for this model' in str_msg
    assert 'BRAF activates MAP2K1.' in str_msg
    # String report given two responses explicitly
    str_msg = qm.make_str_report_one_query(
        'test', query_object, 'pysb', test_response, query_not_appl)
    assert str_msg
    assert 'A new result to query' in str_msg, str_msg
    assert 'Query is not applicable for this model' in str_msg
    assert 'BRAF activates MAP2K1.' in str_msg
    # Html report given two responses explicitly
    html_msg = qm.make_html_one_query_report(
        'test', query_object, 'pysb', test_response, query_not_appl)
    assert html_msg
    assert 'A new result to query' in html_msg
    assert 'Query is not applicable for this model' in html_msg
    assert processed_link in html_msg


@attr('nonpublic')
def test_report_files():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    qm.db.put_queries('tester@test.com', 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_results('test', [(query_object, 'pysb', query_not_appl)])
    results = qm.db.get_results('tester@test.com', latest_order=1)
    qm.make_str_report_per_user(results,
                                filename='test_query_delta.txt')
    report_file = join(dirname(abspath(__file__)), 'test_query_delta.txt')
    with open(report_file, 'r') as f:
        msg = f.read()
    assert msg
    assert 'This is the first result to query' in msg, msg
    assert 'Query is not applicable for this model' in msg
    qm.db.put_results('test', [(query_object, 'pysb', test_response)])
    results = qm.db.get_results('tester@test.com', latest_order=1)
    qm.make_str_report_per_user(results,
                                filename='new_test_query_delta.txt')
    new_report_file = join(dirname(abspath(__file__)),
                           'new_test_query_delta.txt')
    with open(new_report_file, 'r') as f:
        msg = f.read()
    assert msg
    assert 'A new result to query' in msg
    assert 'BRAF activates MAP2K1.' in msg



import logging
from os import path, makedirs, environ
from shutil import copyfile
from configparser import ConfigParser

HERE = path.dirname(path.abspath(__file__))

DB_CONFIG_DIR = path.expanduser('~/.config/emmaa')
DB_CONFIG_PATH = path.join(DB_CONFIG_DIR, 'db_config.ini')
DEFAULT_DB_CONFIG_PATH = path.join(HERE, path.pardir,
                                   'resources/default_db_config.ini')

DB_STR_FMT = "{prefix}://{username}{password}{host}{port}/{name}"
ENV_PREFIX = 'EMMAADB'


logger = logging.getLogger('db_config')


CONFIG_EXISTS = True
if not path.exists(DB_CONFIG_DIR):
    try:
        makedirs(DB_CONFIG_DIR)
    except Exception as e:
        CONFIG_EXISTS = False
        logger.warning("Unable to create config dir: %s" % e)


if not path.exists(DB_CONFIG_PATH) and CONFIG_EXISTS:
    try:
        copyfile(DEFAULT_DB_CONFIG_PATH, DB_CONFIG_PATH)
    except Exception as e:
        CONFIG_EXISTS = False
        logger.warning("Unable to copy config file into config dir: %s" % e)


DATABASES = None


def get_databases(force_update=False, include_config=True):
    global DATABASES
    if DATABASES is None or force_update:
        DATABASES = {}
        if CONFIG_EXISTS and include_config:
            parser = ConfigParser()
            parser.read(DB_CONFIG_PATH)
            for db_name in parser.sections():
                def_dict = {k: parser.get(db_name, k)
                            for k in parser.options(db_name)}
                if def_dict['host']:
                    def_dict['host'] = '@' + def_dict['host']
                def_dict['prefix'] = def_dict['dialect']
                if def_dict['driver']:
                    def_dict['prefix'] += def_dict['driver']
                if def_dict['port']:
                    def_dict['port'] = ':' + def_dict['port']
                if def_dict['password']:
                    def_dict['password'] = ':' + def_dict['password']
                DATABASES[db_name] = DB_STR_FMT.format(**def_dict)
        DATABASES.update({k[len(ENV_PREFIX):].lower(): v
                          for k, v in environ.items()
                          if k.startswith(ENV_PREFIX)})
    return DATABASES



import datetime
from indra.sources import reach
from indra.literature.s3_client import get_reader_json_str, get_full_text
from indra.tools.reading.submit_reading_pipeline import \
    submit_reading, wait_for_complete
from emmaa.statements import EmmaaStatement


def read_pmid_search_terms(pmid_search_terms):
    """Return extracted EmmaaStatements given a PMID-search term dict.

    Parameters
    ----------
    pmid_search_terms : dict
        A dict representing a set of PMIDs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given PMIDs.
    """
    pmids = list(pmid_search_terms.keys())
    date = datetime.datetime.utcnow()
    pmid_stmts = read_pmids(pmids, date)
    estmts = []
    for pmid, stmts in pmid_stmts.items():
        for stmt in stmts:
            es = EmmaaStatement(stmt, date, pmid_search_terms[pmid])
            estmts.append(es)
    return estmts


def read_pmids(pmids, date):
    """Return extracted INDRA Statements per PMID after running reading on AWS.

    Parameters
    ----------
    pmids : list[str]
        A list of PMIDs to read.
    date : datetime
        The date and time associated with the reading, typically the
        current time.

    Returns
    -------
    dict[str, list[indra.statements.Statement]
        A dict of PMIDs and the list of Statements extracted for the given
        PMID by reading.
    """
    date_str = date.strftime('%Y-%m-%d-%H-%M-%S')
    pmid_fname = 'pmids-%s.txt' % date_str
    with open(pmid_fname, 'wt') as fh:
        fh.write('\n'.join(pmids))
    job_list = submit_reading('emmaa', pmid_fname, ['reach'])
    wait_for_complete('run_reach_queue', job_list, idle_log_timeout=600,
                      kill_on_log_timeout=True)
    pmid_stmts = {}
    for pmid in pmids:
        reach_json_str = get_reader_json_str('reach', pmid)
        if reach_json_str is None:
            pmid_stmts[pmid] = []
            continue
        rp = reach.process_json_str(reach_json_str)
        if not rp:
            pmid_stmts[pmid] = []
        else:
            pmid_stmts[pmid] = rp.statements
    return pmid_stmts



import re
import json
import boto3
import logging
import argparse
from urllib import parse
from botocore.exceptions import ClientError
from flask import abort, Flask, request, Response, render_template, jsonify
from flask_jwt_extended import get_jwt_identity, jwt_optional

from indra.statements import get_all_descendants, IncreaseAmount, \
    DecreaseAmount, Activation, Inhibition, AddModification, \
    RemoveModification, get_statement_by_name

from emmaa.util import find_latest_s3_file, strip_out_date, get_s3_client
from emmaa.model import load_config_from_s3
from emmaa.answer_queries import QueryManager, load_model_manager_from_s3
from emmaa.queries import PathProperty, get_agent_from_text, GroundingError

from indralab_auth_tools.auth import auth, config_auth, resolve_auth
from indralab_web_templates.path_templates import path_temps

app = Flask(__name__)
app.register_blueprint(auth)
app.register_blueprint(path_temps)
app.config['DEBUG'] = True
logger = logging.getLogger(__name__)


TITLE = 'emmaa title'
EMMAA_BUCKET_NAME = 'emmaa'
ALL_MODEL_TYPES = ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']
LINKAGE_SYMBOLS = {'LEFT TACK': '\u22a3',
                   'RIGHTWARDS ARROW': '\u2192'}
FORMATTED_MODEL_NAMES = {'pysb': 'PySB',
                         'pybel': 'PyBEL',
                         'signed_graph': 'Signed Graph',
                         'unsigned_graph': 'Unsigned Graph'}
link_list = [('./home', 'EMMAA Dashboard'),
             ('./query', 'Queries')]
SC, jwt = config_auth(app)
qm = QueryManager()


def _get_model_meta_data():
    s3 = boto3.client('s3')
    resp = s3.list_objects(Bucket=EMMAA_BUCKET_NAME, Prefix='models/',
                           Delimiter='/')
    model_data = []
    for pref in resp['CommonPrefixes']:
        model = pref['Prefix'].split('/')[1]
        config_json = get_model_config(model)
        if not config_json:
            continue
        model_data.append((model, config_json))
    return model_data


def get_model_config(model):
    if model in model_cache:
        return model_cache[model]
    try:
        config_json = load_config_from_s3(model)
        model_cache[model] = config_json
    except ClientError:
        logger.warning(f"Model {model} has no metadata. Skipping...")
        return None
    if 'human_readable_name' not in config_json.keys():
        logger.warning(f"Model {model} has no readable name. Skipping...")
        model_cache[model] = None
    return model_cache[model]


def get_model_stats(model, extension='.json'):
    """Gets the latest statistics for the given model

    Parameters
    ----------
    model : str
        Model name to look for
    extension : str

    Returns
    -------
    model_data : json
        The json formatted data containing the statistics for the model
    """
    s3 = get_s3_client()

    # Need jsons for model meta data and test statistics. File name examples:
    # stats/skcm/stats_2019-08-20-17-34-40.json
    prefix = f'stats/{model}/stats_'
    latest_file_key = find_latest_s3_file(bucket=EMMAA_BUCKET_NAME,
                                          prefix=prefix,
                                          extension=extension)
    model_data_object = s3.get_object(Bucket=EMMAA_BUCKET_NAME,
                                      Key=latest_file_key)
    return json.loads(model_data_object['Body'].read().decode('utf8'))


def model_last_updated(model, extension='.pkl'):
    """Find the most recent pickle file of model and return its creation date

    Example file name:
    models/aml/model_2018-12-13-18-11-54.pkl

    Parameters
    ----------
    model : str
        Model name to look for
    extension : str
        The extension the model file needs to have. Default is '.pkl'

    Returns
    -------
    last_updated : str
        A string of the format "YYYY-MM-DD-HH-mm-ss"
    """
    prefix = f'models/{model}/model_'
    try:
        return strip_out_date(find_latest_s3_file(
            bucket=EMMAA_BUCKET_NAME,
            prefix=prefix,
            extension=extension
        ))
    except TypeError:
        logger.info('Could not find latest update date')
        return ''


GLOBAL_PRELOAD = False
model_cache = {}
if GLOBAL_PRELOAD:
    # Load all the model configs
    model_meta_data = _get_model_meta_data()
    # Load all the model managers for queries
    for model, _ in model_meta_data:
        load_model_manager_from_s3(model)


def get_queryable_stmt_types():
    """Return Statement class names that can be used for querying."""
    def get_sorted_descendants(cls):
        return sorted(get_names(get_all_descendants(cls)))

    def get_names(classes):
        return [s.__name__ for s in classes]

    stmt_types = \
        get_names([Activation, Inhibition, IncreaseAmount, DecreaseAmount]) + \
        get_sorted_descendants(AddModification) + \
        get_sorted_descendants(RemoveModification)
    return stmt_types


def _make_query(query_dict, use_grouding_service=True):
    stmt_type = query_dict['typeSelection']
    stmt_class = get_statement_by_name(stmt_type)
    subj = get_agent_from_text(
        query_dict['subjectSelection'], use_grouding_service)
    obj = get_agent_from_text(
        query_dict['objectSelection'], use_grouding_service)
    stmt = stmt_class(subj, obj)
    query = PathProperty(path_stmt=stmt)
    return query


def _extract_stmt_link(anchor_string):
    # Matches an anchor with at least an href attribute
    pattern = '<a.*? href="(.*?)".*?>(.*?)</a>'
    m = re.search(pattern=pattern, string=anchor_string)
    if m:
        return m.group(1), m.group(2)
    else:
        return '', anchor_string


def _new_applied_tests(model_stats_json, model_types, model_name):
    # Extract new applied tests into:
    #   list of tests (one per row)
    #       each test is a list of tuples (one tuple per column)
    #           each tuple is a (href, link_text) pair
    all_test_results = model_stats_json['test_round_summary'][
        'all_test_results']
    new_app_hashes = model_stats_json['tests_delta']['applied_hashes_delta'][
        'added']
    new_app_tests = [(th, all_test_results[th]) for th in new_app_hashes]
    return _format_table_array(new_app_tests, model_types, model_name)


def _format_table_array(tests_json, model_types, model_name):
    # tests_json needs to have the structure: [(test_hash, tests)]
    table_array = []
    for th, test in tests_json:
        new_row = [test['test']]
        for mt in model_types:
            new_row.append((f'/tests/{model_name}/{mt}/{th}', test[mt][0]))

        table_array.append(new_row)
    return table_array


def _new_passed_tests(model_name, model_stats_json, current_model_types):
    new_passed_tests = []
    all_test_results = model_stats_json['test_round_summary'][
        'all_test_results']
    for mt in current_model_types:
        new_passed_hashes = model_stats_json['tests_delta'][mt][
            'passed_hashes_delta']['added']
        if not new_passed_hashes:
            continue
        mt_rows = [[('', f'New passed tests for '
                         f'{FORMATTED_MODEL_NAMES[mt]} model.')]]
        for test_hash in new_passed_hashes:
            test = all_test_results[test_hash]
            path_loc = test[mt][1]
            if isinstance(path_loc, list):
                path = path_loc[0]['path']
            else:
                path = path_loc
            new_row = [(test['test']),
                       (f'/tests/{model_name}/{mt}/{test_hash}', path)]
            mt_rows.append(new_row)
        new_passed_tests += mt_rows
    return new_passed_tests


@app.route('/')
@app.route('/home')
@jwt_optional
def get_home():
    user, roles = resolve_auth(dict(request.args))
    model_data = _get_model_meta_data()
    return render_template('index_template.html', model_data=model_data,
                           link_list=link_list,
                           user_email=user.email if user else "",
                           identity=user.identity() if user else None)


@app.route('/dashboard/<model>')
@jwt_optional
def get_model_dashboard(model):
    user, roles = resolve_auth(dict(request.args))
    model_meta_data = _get_model_meta_data()
    mod_link_list = [('.' + t[0], t[1]) for t in link_list]

    last_update = model_last_updated(model=model)
    ndex_id = 'None available'
    for mid, mmd in model_meta_data:
        if mid == model:
            ndex_id = mmd['ndex']['network']
    if ndex_id == 'None available':
        logger.warning(f'No ndex ID found for {model}')
    if not last_update:
        logger.warning(f'Could not get last update for {model}')
        last_update = 'Not available'
    model_info_contents = [
        [('', 'Last Updated'), ('', last_update)],
        [('', 'Network on Ndex'),
         (f'http://www.ndexbio.org/#/network/{ndex_id}', ndex_id)]]
    model_stats = get_model_stats(model)
    all_new_tests = [(k, v) for k, v in model_stats['test_round_summary'][
        'all_test_results'].items()]
    current_model_types = [mt for mt in ALL_MODEL_TYPES if mt in
                           model_stats['test_round_summary']]
    all_stmts = model_stats['model_summary']['all_stmts']
    most_supported = model_stats['model_summary']['stmts_by_evidence'][:10]
    top_stmts_counts = [
        (all_stmts[h], ('', str(c))) for h, c in most_supported]
    added_stmts_hashes = \
        model_stats['model_delta']['statements_hashes_delta']['added']
    added_stmts = [[(all_stmts[h])] for h in added_stmts_hashes]
    return render_template('model_template.html',
                           model=model,
                           model_data=model_meta_data,
                           model_stats_json=model_stats,
                           link_list=mod_link_list,
                           user_email=user.email if user else "",
                           stmts_counts=top_stmts_counts,
                           added_stmts=added_stmts,
                           model_info_contents=model_info_contents,
                           model_types=["Test", *[FORMATTED_MODEL_NAMES[mt]
                                                  for mt in 
                                                  current_model_types]],
                           new_applied_tests=_new_applied_tests(
                               model_stats_json=model_stats,
                               model_types=current_model_types,
                               model_name=model),
                           all_test_results=_format_table_array(
                               tests_json=all_new_tests,
                               model_types=current_model_types,
                               model_name=model),
                           new_passed_tests=_new_passed_tests(
                               model, model_stats, current_model_types))


@app.route('/tests/<model>/<model_type>/<test_hash>')
def get_model_tests_page(model, model_type, test_hash):
    if model_type not in ALL_MODEL_TYPES:
        abort(Response(f'Model type {model_type} does not exist', 404))
    model_meta_data = _get_model_meta_data()
    mod_link_list = [('../../../.' + t[0], t[1]) for t in link_list]
    ndex_id = 'None available'
    for mid, mmd in model_meta_data:
        if mid == model:
            ndex_id = mmd['ndex']['network']
    if ndex_id == 'None available':
        logger.warning(f'No ndex ID found for {model}')
    model_stats = get_model_stats(model)
    current_test = \
        model_stats['test_round_summary']['all_test_results'][test_hash]
    current_model_types = [mt for mt in ALL_MODEL_TYPES if mt in
                           model_stats['test_round_summary']]
    test = current_test["test"]
    test_status, path_list = current_test[model_type]
    return render_template('tests_template.html',
                           link_list=mod_link_list,
                           model=model,
                           model_type=model_type,
                           all_model_types=current_model_types,
                           test_hash=test_hash,
                           model_stats_json=model_stats,
                           ndexID=ndex_id,
                           test=test,
                           test_status=test_status,
                           path_list=path_list,
                           formatted_names=FORMATTED_MODEL_NAMES)


@app.route('/query')
@jwt_optional
def get_query_page():
    user, roles = resolve_auth(dict(request.args))
    user_email = user.email if user else ""
    user_id = user.id if user else None
    model_meta_data = _get_model_meta_data()
    stmt_types = get_queryable_stmt_types()

    # user_email = 'joshua@emmaa.com'
    old_results = qm.get_registered_queries(user_email) if user_email else []

    return render_template('query_template.html', model_data=model_meta_data,
                           stmt_types=stmt_types, old_results=old_results,
                           link_list=link_list, user_email=user_email,
                           user_id=user_id, model_names=FORMATTED_MODEL_NAMES)


@app.route('/query/submit', methods=['POST'])
@jwt_optional
def process_query():
    # Print inputs.
    logger.info('Got model query')
    logger.info("Args -----------")
    logger.info(request.args)
    logger.info("Json -----------")
    logger.info(str(request.json))
    logger.info("------------------")

    user, roles = resolve_auth(dict(request.args))
    user_email = user.email if user else ""
    user_id = user.id if user else None

    # Extract info.
    expected_query_keys = {f'{pos}Selection'
                           for pos in ['subject', 'object', 'type']}
    expected_models = {mid for mid, _ in _get_model_meta_data()}
    try:
        # If user tries to register query without logging in, refuse query
        # with 401 (unauthorized)
        if request.json['register']:
            if user_email:
                # Logged in
                subscribe = request.json['register']
            else:
                # Not logged in
                logger.warning('User not logged in! Query will not be '
                               'registered.')
                return jsonify({'result': 'failure',
                                'reason': 'Invalid credentials'}), 401
        # Does not try to register
        else:
            subscribe = False
        query_json = request.json['query']
        assert set(query_json.keys()) == expected_query_keys, \
            (f'Did not get expected query keys: got {set(query_json.keys())} '
             f'not {expected_query_keys}')
        models = set(request.json.get('models'))
        assert models < expected_models, \
            f'Got unexpected models: {models - expected_models}'
    except (KeyError, AssertionError) as e:
        logger.exception(e)
        logger.error("Invalid query!")
        abort(Response(f'Invalid request: {str(e)}', 400))
    try:
        query = _make_query(query_json)
    except GroundingError as e:
        logger.exception(e)
        logger.error("Invalid grounding!")
        abort(Response(f'Invalid entity: {str(e)}', 400))

    is_test = 'test' in request.json or 'test' == request.json.get('tag')

    if is_test:
        logger.info('Test passed')
        res = {'result': 'test passed', 'ref': None}

    else:
        logger.info('Query submitted')
        try:
            result = qm.answer_immediate_query(
                user_email, user_id, query, models, subscribe)
        except Exception as e:
            logger.exception(e)
            raise(e)
        logger.info('Answer to query received, responding to client.')
        res = {'result': result}

    logger.info('Result: %s' % str(res))
    return Response(json.dumps(res), mimetype='application/json')


if __name__ == '__main__':
    parser = argparse.ArgumentParser('Run the EMMAA dashboard service.')
    parser.add_argument('--host', default='0.0.0.0')
    parser.add_argument('--port', default=5000, type=int)
    parser.add_argument('--preload', action='store_true')
    args = parser.parse_args()

    # TODO: make pre-loading available when running service via Gunicorn
    if args.preload and not GLOBAL_PRELOAD:
        # Load all the model configs
        model_meta_data = _get_model_meta_data()
        # Load all the model mamangers for queries
        for model, _ in model_meta_data:
            load_model_manager_from_s3(model)

    print(app.url_map)  # Get all avilable urls and link them
    app.run(host=args.host, port=args.port)



import re
import logging

from .config import get_databases
from .manager import EmmaaDatabaseManager

logger = logging.getLogger(__name__)


def get_db(name):
    """Get a db instance based on its name in the config or env."""
    defaults = get_databases()
    db_name = defaults[name]
    m = re.match('(\w+)://.*?/([\w.]+)', db_name)
    if m is None:
        logger.error("Poorly formed db name: %s" % db_name)
        return
    return EmmaaDatabaseManager(db_name, label=name)



__all__ = ['User', 'Query', 'UserQuery', 'Result']

import logging

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, UniqueConstraint, ForeignKey, \
    Boolean, DateTime, func, BigInteger
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import JSONB


logger = logging.getLogger(__name__)


Base = declarative_base()


class EmmaaTable(object):
    _skip_disp = []

    def _content_strings(self):
        ret = []
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                if k in self._skip_disp:
                    ret.append(f'{k}=[not shown]')
                else:
                    ret.append(f'{k}={v}')
        return ret

    def _make_str(self):
        s = self.__tablename__ + ':\n'
        s += '\n'.join(f'\t{line}' for line in self._content_strings())
        return s

    def display(self):
        """Display the values of this entry."""
        print(self._make_str())

    def __str__(self):
        return self._make_str()

    def __repr__(self):
        return f'{self.__class__.__name__}' \
               f'({", ".join(self._content_strings())})'


class User(Base, EmmaaTable):
    """A table containing users of EMMAA: ``User(_id_, email)``

    Parameters
    ----------
    id : int
        (from indralab_auth_tools.src.models.User.id, primary key)
        A database-generated integer from the User table in indralab
        auth tools.
    email : str
        The email of the user (must be unique)
    """
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)
    email = Column(String, unique=True)


class Query(Base, EmmaaTable):
    """Queries run on each model: ``Query(_hash_, model_id, json)``

    The hash column is a hash generated from the json and model_id columns
    that can be derived from the

    Parameters
    ----------
    hash : big-int
        (primary key) A 32 bit integer generated from the json and model_id.
    model_id : str
        (20 character) The short id/acronym for the given model.
    json : json
        A json dict containing the relevant parameters defining the query.
    """
    __tablename__ = 'query'
    hash = Column(BigInteger, primary_key=True)
    model_id = Column(String(20), nullable=False)
    json = Column(JSONB, nullable=False)
    __table_args__ = (
        UniqueConstraint('model_id', 'json', name='query-uniqueness'),
        )


class UserQuery(Base, EmmaaTable):
    """A table linking users to queries:

    ``UserQuery(_id_, user_id, query_hash, date, subscription)``

    Parameters
    ----------
    id : int
        (auto, primary key) A database-assigned integer id.
    user_id : int
        (foreign key -> User.id) The id of the user related to this query.
    query_hash : big-int
        (foreign key -> Query.hash) The hash of the query json, which can be
        directly generated.
    date : datetime
        (auto) The date that this entry was added to the database.
    subscription : bool
        Record whether the user has subscribed to see results of this model.
    count : int
        Record the number of times the user associated with user id has done
        this query
    """
    __tablename__ = 'user_query'
    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=True)
    user = relationship(User)
    query_hash = Column(BigInteger, ForeignKey('query.hash'), nullable=False)
    query = relationship(Query)
    date = Column(DateTime, default=func.now())
    subscription = Column(Boolean, nullable=False)
    count = Column(Integer, nullable=False)


class Result(Base, EmmaaTable):
    """Results of queries to models:

    ``Result(_id_, query_hash, date, result_json)``

    Parameters
    ----------
    id : int
        (auto, primary key) A database-assigned integer id.
    query_hash : big-int
        (foreign key -> Query.hash) The hash of the query json, which can be
        directly generated.
    date : datetime
        (auto) The date the result was entered into the database.
    result_json : json
        A json dict containing the results for the query.
    mc_type : str
        A name of a ModelChecker used to answer the query.
    """
    __tablename__ = 'result'
    id = Column(Integer, primary_key=True)
    query_hash = Column(BigInteger, ForeignKey('query.hash'), nullable=False)
    query = relationship(Query)
    date = Column(DateTime, default=func.now())
    result_json = Column(JSONB, nullable=False)
    mc_type = Column(String(20), default='pysb')
